/*
 ============================================================================
 Name        : $(baseName).cu
 Author      : $(author)
 Version     :
 Copyright   : $(copyright)
 Description : CUDA compute and verify SAXPY
 ============================================================================
 */
 
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/* Prototypes */
bool isZero(float x) { return abs(x)<1e-6;}
float randomFloat(void);
float *safeFloatMalloc(unsigned);
static void CheckCudaErrorAux (const char *, unsigned, const char *, cudaError_t);
#define CUDA_CHECK_RETURN(value) CheckCudaErrorAux(__FILE__,__LINE__, #value, value)

/**
 * CUDA SAXPY kernel
 */
__global__ void saxpyKernel(unsigned n, float a, const float *x, float *y)
{
	unsigned idx = blockIdx.x*blockDim.x+threadIdx.x;
	if (idx < n)
		y[idx] += a*x[idx];
}

/**
 * Routine that tests SAXPY kernel, by allocating two device accessible vectors,
 * initializing them, launching saxpy kernel and later verifying results
 */
int testSAXPY(int vectorSize, const float a)
{
	unsigned i;
	float *x,*y,*y_copy;
	float *gpuX, *gpuY;
	static const int BLOCK_SIZE = 1024;
	int blockCount = (vectorSize+BLOCK_SIZE-1)/BLOCK_SIZE;
	int rc = 0;

	/* Initialize vectors with random values */

	x = safeFloatMalloc(vectorSize);
	y = safeFloatMalloc(vectorSize);
	y_copy = safeFloatMalloc(vectorSize);
	CUDA_CHECK_RETURN(cudaMalloc(&gpuX,sizeof(float)*vectorSize));
	CUDA_CHECK_RETURN(cudaMalloc(&gpuY,sizeof(float)*vectorSize));
	for (i=0; i < vectorSize; ++i) {
		x[i] = randomFloat();
		y_copy[i] = y[i] = randomFloat();
	}

	/* Copy memory to GPU */
	CUDA_CHECK_RETURN(cudaMemcpy(gpuX, x, sizeof(float)*vectorSize, cudaMemcpyHostToDevice));
	CUDA_CHECK_RETURN(cudaMemcpy(gpuY, y, sizeof(float)*vectorSize, cudaMemcpyHostToDevice));
	/* Launch SAXPY on GPU */
	saxpyKernel<<<blockCount,BLOCK_SIZE>>>(vectorSize, a, gpuX, gpuY);
	/* Wait for kernel to finish and copy results back */
	CUDA_CHECK_RETURN(cudaMemcpy(y, gpuY, sizeof(float)*vectorSize, cudaMemcpyDeviceToHost));

	/* Verify results */
	for (i = 0; i < vectorSize; ++i)
		if (!isZero(y_copy[i]+a*x[i] - y[i])) {
			fprintf(stderr, "%u: %g+%g*%g != %g\n", i, y_copy[i], a, x[i], y[i]);
			rc = -1;
			break;
		}
	if (rc == 0)
		printf("Test passes\n");

	/* Free memory and return results */
	CUDA_CHECK_RETURN(cudaFree(gpuX));
	CUDA_CHECK_RETURN(cudaFree(gpuY));
	free (x);
	free (y);
	free (y_copy);
	return rc;
}

int main(void)
{
	return testSAXPY(1<<20, .4);
}

/**
 * Return non-negative floating point value in range from 0 to 1.0
 */
float randomFloat(void)
{
	static const float INV_RANDMAX = 1.0/RAND_MAX;
	return INV_RANDMAX*rand();
}

/**
 * Allocate an array of nElements
 * Abort the application execution of allocation failed
 */
float *safeFloatMalloc(unsigned nElements)
{
	float *rc = (float *) malloc (sizeof (float) * nElements);
	if (rc == NULL) {
		fprintf (stderr, "Failed to allocate %u bytes\n", (unsigned)sizeof (float) * nElements);
		exit (1);
	}
	return rc;
}

/**
 * Check the return value of the CUDA runtime API call and exit
 * the application if the call has failed.
 */
static void CheckCudaErrorAux (const char *file, unsigned line, const char *statement, cudaError_t err)
{
	if (err == cudaSuccess)
		return;
	fprintf (stderr, "%s returned %s (%u) at %s:%u\n", statement, cudaGetErrorString(err), err, file, line);
	exit (1);
}
